@@ -20,7 +20,9 @@ from ray.rllib.rollout import rollout
 from ray.rllib.tests.test_external_env import SimpleServing
 from ray.tune.registry import register_env
 from ray.rllib.utils.framework import try_import_tf, try_import_torch
+from ray.rllib.utils.numpy import one_hot
 from ray.rllib.utils.spaces.repeated import Repeated
+from ray.rllib.utils.test_utils import check
 
 tf1, tf, tfv = try_import_tf()
 _, nn = try_import_torch()
@@ -69,12 +71,6 @@ REPEATED_SPACE = Repeated(PLAYER_SPACE, max_len=MAX_PLAYERS)
 REPEATED_SAMPLES = [REPEATED_SPACE.sample() for _ in range(10)]
 
 
-def one_hot(i, n):
-    out = [0.0] * n
-    out[i] = 1.0
-    return out
-
-
 class NestedDictEnv(gym.Env):
     def __init__(self):
         self.action_space = spaces.Discrete(2)
@@ -234,6 +230,17 @@ def to_list(value):
 class DictSpyModel(TFModelV2):
     capture_index = 0
 
+    def __init__(self, obs_space, action_space, num_outputs, model_config,
+                 name):
+        super().__init__(obs_space, action_space, None, model_config, name)
+        # Will only feed in sensors->pos.
+        input_ = tf.keras.layers.Input(
+            shape=self.obs_space.original_space["sensors"]["position"].shape)
+
+        self.num_outputs = num_outputs or 64
+        out = tf.keras.layers.Dense(self.num_outputs)(input_)
+        self._main_layer = tf.keras.models.Model([input_], [out])
+
     def forward(self, input_dict, state, seq_lens):
         def spy(pos, front_cam, task):
             # TF runs this function in an isolated context, so we have to use
@@ -255,14 +262,26 @@ class DictSpyModel(TFModelV2):
             stateful=True)
 
         with tf1.control_dependencies([spy_fn]):
-            output = tf1.layers.dense(input_dict["obs"]["sensors"]["position"],
-                                      self.num_outputs)
+            output = self._main_layer(
+                [input_dict["obs"]["sensors"]["position"]])
+
         return output, []
 
 
 class TupleSpyModel(TFModelV2):
     capture_index = 0
 
+    def __init__(self, obs_space, action_space, num_outputs, model_config,
+                 name):
+        super().__init__(obs_space, action_space, None, model_config, name)
+        # Will only feed in 0th index of observation Tuple space.
+        input_ = tf.keras.layers.Input(
+            shape=self.obs_space.original_space[0].shape)
+
+        self.num_outputs = num_outputs or 64
+        out = tf.keras.layers.Dense(self.num_outputs)(input_)
+        self._main_layer = tf.keras.models.Model([input_], [out])
+
     def forward(self, input_dict, state, seq_lens):
         def spy(pos, cam, task):
             # TF runs this function in an isolated context, so we have to use
@@ -338,6 +357,9 @@ class NestedSpacesTest(unittest.TestCase):
                 },
                 "framework": "tf",
             })
+        # Skip first passes as they came from the TorchPolicy loss
+        # initialization.
+        DictSpyModel.capture_index = 0
         pg.train()
 
         # Check that the model sees the correct reconstructed observations
@@ -351,7 +373,7 @@ class NestedSpacesTest(unittest.TestCase):
                 DICT_SAMPLES[i]["inner_state"]["job_status"]["task"], 5)
             self.assertEqual(seen[0][0].tolist(), pos_i)
             self.assertEqual(seen[1][0].tolist(), cam_i)
-            self.assertEqual(seen[2][0].tolist(), task_i)
+            check(seen[2][0], task_i)
 
     def do_test_nested_tuple(self, make_env):
         ModelCatalog.register_custom_model("composite2", TupleSpyModel)
@@ -367,6 +389,9 @@ class NestedSpacesTest(unittest.TestCase):
                 },
                 "framework": "tf",
             })
+        # Skip first passes as they came from the TorchPolicy loss
+        # initialization.
+        TupleSpyModel.capture_index = 0
         pg.train()
 
         # Check that the model sees the correct reconstructed observations
@@ -379,7 +404,7 @@ class NestedSpacesTest(unittest.TestCase):
             task_i = one_hot(TUPLE_SAMPLES[i][2], 5)
             self.assertEqual(seen[0][0].tolist(), pos_i)
             self.assertEqual(seen[1][0].tolist(), cam_i)
-            self.assertEqual(seen[2][0].tolist(), task_i)
+            check(seen[2][0], task_i)
 
     def test_nested_dict_gym(self):
         self.do_test_nested_dict(lambda _: NestedDictEnv())
@@ -438,6 +463,9 @@ class NestedSpacesTest(unittest.TestCase):
                 },
                 "framework": "tf",
             })
+        # Skip first passes as they came from the TorchPolicy loss
+        # initialization.
+        TupleSpyModel.capture_index = DictSpyModel.capture_index = 0
         pg.train()
 
         for i in range(4):
@@ -450,7 +478,7 @@ class NestedSpacesTest(unittest.TestCase):
                 DICT_SAMPLES[i]["inner_state"]["job_status"]["task"], 5)
             self.assertEqual(seen[0][0].tolist(), pos_i)
             self.assertEqual(seen[1][0].tolist(), cam_i)
-            self.assertEqual(seen[2][0].tolist(), task_i)
+            check(seen[2][0], task_i)
 
         for i in range(4):
             seen = pickle.loads(
@@ -461,7 +489,7 @@ class NestedSpacesTest(unittest.TestCase):
             task_i = one_hot(TUPLE_SAMPLES[i][2], 5)
             self.assertEqual(seen[0][0].tolist(), pos_i)
             self.assertEqual(seen[1][0].tolist(), cam_i)
-            self.assertEqual(seen[2][0].tolist(), task_i)
+            check(seen[2][0], task_i)
 
     def test_rollout_dict_space(self):
         register_env("nested", lambda _: NestedDictEnv())
@@ -493,6 +521,9 @@ class NestedSpacesTest(unittest.TestCase):
                 "framework": "torch",
             })
 
+        # Skip first passes as they came from the TorchPolicy loss
+        # initialization.
+        TorchSpyModel.capture_index = 0
         a2c.train()
 
         # Check that the model sees the correct reconstructed observations
@@ -500,13 +531,16 @@ class NestedSpacesTest(unittest.TestCase):
             seen = pickle.loads(
                 ray.experimental.internal_kv._internal_kv_get(
                     "torch_spy_in_{}".format(i)))
+
             pos_i = DICT_SAMPLES[i]["sensors"]["position"].tolist()
             cam_i = DICT_SAMPLES[i]["sensors"]["front_cam"][0].tolist()
             task_i = one_hot(
                 DICT_SAMPLES[i]["inner_state"]["job_status"]["task"], 5)
-            self.assertEqual(seen[0][0].tolist(), pos_i)
-            self.assertEqual(seen[1][0].tolist(), cam_i)
-            self.assertEqual(seen[2][0].tolist(), task_i)
+            # Only look at the last entry (-1) in `seen` as we reset (re-use)
+            # the ray-kv indices before training.
+            self.assertEqual(seen[0][-1].tolist(), pos_i)
+            self.assertEqual(seen[1][-1].tolist(), cam_i)
+            check(seen[2][-1], task_i)
 
     # TODO(ekl) should probably also add a test for TF/eager
     def test_torch_repeated(self):
@@ -524,6 +558,9 @@ class NestedSpacesTest(unittest.TestCase):
                 "framework": "torch",
             })
 
+        # Skip first passes as they came from the TorchPolicy loss
+        # initialization.
+        TorchRepeatedSpyModel.capture_index = 0
         a2c.train()
 
         # Check that the model sees the correct reconstructed observations
@@ -531,7 +568,11 @@ class NestedSpacesTest(unittest.TestCase):
             seen = pickle.loads(
                 ray.experimental.internal_kv._internal_kv_get(
                     "torch_rspy_in_{}".format(i)))
-            self.assertEqual(to_list(seen), [to_list(REPEATED_SAMPLES[i])])
+
+            # Only look at the last entry (-1) in `seen` as we reset (re-use)
+            # the ray-kv indices before training.
+            self.assertEqual(
+                to_list(seen[:][-1]), to_list(REPEATED_SAMPLES[i]))
 
 
 if __name__ == "__main__":
