@@ -1,26 +1,57 @@
+import asyncio
 import atexit
+import time
 from functools import wraps
 import os
+from uuid import UUID
+import threading
+from typing import Any, Callable, Coroutine, Dict, List, Optional, Type, Union
+from dataclasses import dataclass
+from warnings import warn
 
 import ray
 from ray.serve.constants import (DEFAULT_HTTP_HOST, DEFAULT_HTTP_PORT,
                                  SERVE_CONTROLLER_NAME, HTTP_PROXY_TIMEOUT)
-from ray.serve.controller import ServeController
-from ray.serve.handle import RayServeHandle
+from ray.serve.controller import ServeController, BackendTag, ReplicaTag
+from ray.serve.handle import RayServeHandle, RayServeSyncHandle
 from ray.serve.utils import (block_until_http_ready, format_actor_name,
-                             get_random_letters, logger, get_conda_env_dir)
+                             get_random_letters, logger, get_conda_env_dir,
+                             get_current_node_resource_key)
 from ray.serve.exceptions import RayServeException
-from ray.serve.config import BackendConfig, ReplicaConfig, BackendMetadata
+from ray.serve.config import (BackendConfig, ReplicaConfig, BackendMetadata,
+                              HTTPOptions)
 from ray.serve.env import CondaEnv
+from ray.serve.router import RequestMetadata, Router
 from ray.actor import ActorHandle
-from typing import Any, Callable, Dict, List, Optional, Type, Union
 
-_INTERNAL_CONTROLLER_NAME = None
+_INTERNAL_REPLICA_CONTEXT = None
+global_async_loop = None
 
 
-def _set_internal_controller_name(name):
-    global _INTERNAL_CONTROLLER_NAME
-    _INTERNAL_CONTROLLER_NAME = name
+@dataclass
+class InternalReplicaContext:
+    """Stores data for Serve API calls from within the user's backend code."""
+    backend_tag: BackendTag
+    replica_tag: ReplicaTag
+    controller_name: str
+
+
+def create_or_get_async_loop_in_thread():
+    global global_async_loop
+    if global_async_loop is None:
+        global_async_loop = asyncio.new_event_loop()
+        thread = threading.Thread(
+            daemon=True,
+            target=global_async_loop.run_forever,
+        )
+        thread.start()
+    return global_async_loop
+
+
+def _set_internal_replica_context(backend_tag, replica_tag, controller_name):
+    global _INTERNAL_REPLICA_CONTEXT
+    _INTERNAL_REPLICA_CONTEXT = InternalReplicaContext(
+        backend_tag, replica_tag, controller_name)
 
 
 def _ensure_connected(f: Callable) -> Callable:
@@ -33,6 +64,35 @@ def _ensure_connected(f: Callable) -> Callable:
     return check
 
 
+class ThreadProxiedRouter:
+    def __init__(self, controller_handle, sync: bool):
+        self.router = Router(controller_handle)
+
+        if sync:
+            self.async_loop = create_or_get_async_loop_in_thread()
+            asyncio.run_coroutine_threadsafe(
+                self.router.setup_in_async_loop(),
+                self.async_loop,
+            )
+        else:
+            self.async_loop = asyncio.get_event_loop()
+            self.async_loop.create_task(self.router.setup_in_async_loop())
+
+    def _remote(self, endpoint_name, handle_options, request_data,
+                kwargs) -> Coroutine:
+        request_metadata = RequestMetadata(
+            get_random_letters(10),  # Used for debugging.
+            endpoint_name,
+            call_method=handle_options.method_name,
+            shard_key=handle_options.shard_key,
+            http_method=handle_options.http_method,
+            http_headers=handle_options.http_headers,
+        )
+        coro = self.router.assign_request(request_metadata, request_data,
+                                          **kwargs)
+        return coro
+
+
 class Client:
     def __init__(self,
                  controller: ActorHandle,
@@ -42,13 +102,10 @@ class Client:
         self._controller_name = controller_name
         self._detached = detached
         self._shutdown = False
+        self._http_config = ray.get(controller.get_http_config.remote())
 
-        # NOTE(simon): Used to cache client.get_handle(endpoint) call. It will
-        # mostly grow in size, it will only shrink when user calls the
-        # .remove_endpoint method. This is fine because we expect the number of
-        # endpoints to be fairly small. However, in case this dictionary does
-        # grow very big, we can replace it with a LRU cache instead.
-        self._handle_cache: Dict[str, ActorHandle] = dict()
+        self._sync_proxied_router = None
+        self._async_proxied_router = None
 
         # NOTE(edoakes): Need this because the shutdown order isn't guaranteed
         # when the interpreter is exiting so we can't rely on __del__ (it
@@ -60,11 +117,23 @@ class Client:
 
             atexit.register(shutdown_serve_client)
 
+    def _get_proxied_router(self, sync: bool):
+        if sync:
+            if self._sync_proxied_router is None:
+                self._sync_proxied_router = ThreadProxiedRouter(
+                    self._controller, sync=True)
+            return self._sync_proxied_router
+        else:
+            if self._async_proxied_router is None:
+                self._async_proxied_router = ThreadProxiedRouter(
+                    self._controller, sync=False)
+            return self._async_proxied_router
+
     def __del__(self):
         if not self._detached:
-            logger.info("Shutting down Ray Serve because client went out of "
-                        "scope. To prevent this, either keep a reference to "
-                        "the client object or use serve.start(detached=True).")
+            logger.debug("Shutting down Ray Serve because client went out of "
+                         "scope. To prevent this, either keep a reference to "
+                         "the client or use serve.start(detached=True).")
             self.shutdown()
 
     def __reduce__(self):
@@ -78,11 +147,34 @@ class Client:
         Shuts down all processes and deletes all state associated with the
         instance.
         """
-        if not self._shutdown:
+        if (not self._shutdown) and ray.is_initialized():
             ray.get(self._controller.shutdown.remote())
             ray.kill(self._controller, no_restart=True)
+
+            # Wait for the named actor entry gets removed as well.
+            started = time.time()
+            while True:
+                try:
+                    ray.get_actor(self._controller_name)
+                    if time.time() - started > 5:
+                        logger.warning(
+                            "Waited 5s for Serve to shutdown gracefully but "
+                            "the controller is still not cleaned up. "
+                            "You can ignore this warning if you are shutting "
+                            "down the Ray cluster.")
+                        break
+                except ValueError:  # actor name is removed
+                    break
+
             self._shutdown = True
 
+    @_ensure_connected
+    def _wait_for_goal(self, result_object_id: ray.ObjectRef) -> bool:
+        goal_id: Optional[UUID] = ray.get(result_object_id)
+        if goal_id is not None:
+            ray.get(self._controller.wait_for_goal.remote(goal_id))
+            logger.debug(f"Goal {goal_id} completed.")
+
     @_ensure_connected
     def create_endpoint(self,
                         endpoint_name: str,
@@ -137,18 +229,39 @@ class Client:
                     "an element of type {}".format(type(method)))
             upper_methods.append(method.upper())
 
-        ray.get(
+        self._wait_for_goal(
             self._controller.create_endpoint.remote(
                 endpoint_name, {backend: 1.0}, route, upper_methods))
 
+        # Block until the route table has been propagated to all HTTP proxies.
+        if route is not None:
+
+            def check_ready(http_response):
+                return route in http_response.json()
+
+            futures = []
+            for node_id in ray.state.node_ids():
+                future = block_until_http_ready.options(
+                    num_cpus=0, resources={
+                        node_id: 0.01
+                    }).remote(
+                        "http://{}:{}/-/routes".format(self._http_config.host,
+                                                       self._http_config.port),
+                        check_ready=check_ready,
+                        timeout=HTTP_PROXY_TIMEOUT)
+                futures.append(future)
+            try:
+                ray.get(futures)
+            except ray.exceptions.RayTaskError:
+                raise TimeoutError("Route not available at HTTP proxies "
+                                   "after {HTTP_PROXY_TIMEOUT}s.")
+
     @_ensure_connected
     def delete_endpoint(self, endpoint: str) -> None:
         """Delete the given endpoint.
 
         Does not delete any associated backends.
         """
-        if endpoint in self._handle_cache:
-            del self._handle_cache[endpoint]
         ray.get(self._controller.delete_endpoint.remote(endpoint))
 
     @_ensure_connected
@@ -193,7 +306,7 @@ class Client:
                 "config_options must be a BackendConfig or dictionary.")
         if isinstance(config_options, dict):
             config_options = BackendConfig.parse_obj(config_options)
-        ray.get(
+        self._wait_for_goal(
             self._controller.update_backend_config.remote(
                 backend_tag, config_options))
 
@@ -222,8 +335,9 @@ class Client:
         Args:
             backend_tag (str): a unique tag assign to identify this backend.
             func_or_class (callable, class): a function or a class implementing
-                __call__.
-            actor_init_args (optional): the arguments to pass to the class.
+                __call__, returning a JSON-serializable object or a
+                Starlette Response object.
+            *actor_init_args (optional): the arguments to pass to the class
                 initialization method.
             ray_actor_options (optional): options to be passed into the
                 @ray.remote decorator for the backend actor.
@@ -290,7 +404,7 @@ class Client:
             raise TypeError("config must be a BackendConfig or a dictionary.")
 
         backend_config._validate_complete()
-        ray.get(
+        self._wait_for_goal(
             self._controller.create_backend.remote(backend_tag, backend_config,
                                                    replica_config))
 
@@ -303,12 +417,18 @@ class Client:
         return ray.get(self._controller.get_all_backends.remote())
 
     @_ensure_connected
-    def delete_backend(self, backend_tag: str) -> None:
+    def delete_backend(self, backend_tag: str, force: bool = False) -> None:
         """Delete the given backend.
 
         The backend must not currently be used by any endpoints.
+
+        Args:
+            backend_tag (str): The backend tag to be deleted.
+            force (bool): Whether or not to force the deletion, without waiting
+              for graceful shutdown. Default to false.
         """
-        ray.get(self._controller.delete_backend.remote(backend_tag))
+        self._wait_for_goal(
+            self._controller.delete_backend.remote(backend_tag, force))
 
     @_ensure_connected
     def set_traffic(self, endpoint_name: str,
@@ -358,15 +478,20 @@ class Client:
                                                    proportion))
 
     @_ensure_connected
-    def get_handle(self,
-                   endpoint_name: str,
-                   missing_ok: Optional[bool] = False) -> RayServeHandle:
+    def get_handle(
+            self,
+            endpoint_name: str,
+            missing_ok: Optional[bool] = False,
+            sync: bool = True) -> Union[RayServeHandle, RayServeSyncHandle]:
         """Retrieve RayServeHandle for service endpoint to invoke it from Python.
 
         Args:
             endpoint_name (str): A registered service endpoint.
             missing_ok (bool): If true, then Serve won't check the endpoint is
                 registered. False by default.
+            sync (bool): If true, then Serve will return a ServeHandle that
+                works everywhere. Otherwise, Serve will return a ServeHandle
+                that's only usable in asyncio loop.
 
         Returns:
             RayServeHandle
@@ -375,16 +500,37 @@ class Client:
                 self._controller.get_all_endpoints.remote()):
             raise KeyError(f"Endpoint '{endpoint_name}' does not exist.")
 
-        if endpoint_name not in self._handle_cache:
-            handle = RayServeHandle(self._controller, endpoint_name, sync=True)
-            self._handle_cache[endpoint_name] = handle
-        return self._handle_cache[endpoint_name]
+        if asyncio.get_event_loop().is_running() and sync:
+            logger.warning(
+                "You are retrieving a sync handle inside an asyncio loop. "
+                "Try getting client.get_handle(.., sync=False) to get better "
+                "performance. Learn more at https://docs.ray.io/en/master/"
+                "serve/advanced.html#sync-and-async-handles")
+
+        if not asyncio.get_event_loop().is_running() and not sync:
+            logger.warning(
+                "You are retrieving an async handle outside an asyncio loop. "
+                "You should make sure client.get_handle is called inside a "
+                "running event loop. Or call client.get_handle(.., sync=True) "
+                "to create sync handle. Learn more at https://docs.ray.io/en/"
+                "master/serve/advanced.html#sync-and-async-handles")
+
+        if sync:
+            handle = RayServeSyncHandle(
+                self._get_proxied_router(sync=sync), endpoint_name)
+        else:
+            handle = RayServeHandle(
+                self._get_proxied_router(sync=sync), endpoint_name)
+        return handle
 
 
-def start(detached: bool = False,
-          http_host: str = DEFAULT_HTTP_HOST,
-          http_port: int = DEFAULT_HTTP_PORT,
-          http_middlewares: List[Any] = []) -> Client:
+def start(
+        detached: bool = False,
+        http_host: Optional[str] = DEFAULT_HTTP_HOST,
+        http_port: int = DEFAULT_HTTP_PORT,
+        http_middlewares: List[Any] = [],
+        http_options: Optional[Union[dict, HTTPOptions]] = None,
+) -> Client:
     """Initialize a serve instance.
 
     By default, the instance will be scoped to the lifetime of the returned
@@ -395,15 +541,44 @@ def start(detached: bool = False,
 
     Args:
         detached (bool): Whether not the instance should be detached from this
-            script.
-        http_host (str): Host for HTTP servers to listen on. Defaults to
-            "127.0.0.1". To expose Serve publicly, you probably want to set
-            this to "0.0.0.0". One HTTP server will be started on each node in
-            the Ray cluster. To not start HTTP servers, set this to None.
-        http_port (int): Port for HTTP server. Defaults to 8000.
-        http_middlewares (list): A list of Starlette middlewares that will be
-            applied to the HTTP servers in the cluster.
+          script.
+        http_host (Optional[str]): Deprecated, use http_options instead.
+        http_port (int): Deprecated, use http_options instead.
+        http_middlewares (list): Deprecated, use http_options instead.
+        http_options (Optional[Dict, serve.HTTPOptions]): Configuration options
+          for HTTP proxy. You can pass in a dictionary or HTTPOptions object
+          with fields:
+
+            - host(str, None): Host for HTTP servers to listen on. Defaults to
+              "127.0.0.1". To expose Serve publicly, you probably want to set
+              this to "0.0.0.0".
+            - port(int): Port for HTTP server. Defaults to 8000.
+            - middlewares(list): A list of Starlette middlewares that will be
+              applied to the HTTP servers in the cluster.
+            - location(str, serve.config.DeploymentMode): The deployment
+              location of HTTP servers:
+
+                - "HeadOnly": start one HTTP server on the head node. Serve
+                  assumes the head node is the node you executed serve.start
+                  on. This is the default.
+                - "EveryNode": start one HTTP server per node.
+                - "NoServer" or None: disable HTTP server.
     """
+    if ((http_host != DEFAULT_HTTP_HOST) or (http_port != DEFAULT_HTTP_PORT)
+            or (len(http_middlewares) != 0)):
+        if http_options is not None:
+            raise ValueError(
+                "You cannot specify both `http_options` and any of the "
+                "`http_host`, `http_port`, and `http_middlewares` arguments. "
+                "`http_options` is preferred.")
+        else:
+            warn(
+                "`http_host`, `http_port`, `http_middlewares` are deprecated. "
+                "Please use serve.start(http_options={'host': ..., "
+                "'port': ..., middlewares': ...}) instead.",
+                DeprecationWarning,
+            )
+
     # Initialize ray if needed.
     if not ray.is_initialized():
         ray.init()
@@ -423,29 +598,37 @@ def start(detached: bool = False,
         controller_name = format_actor_name(SERVE_CONTROLLER_NAME,
                                             get_random_letters())
 
+    if isinstance(http_options, dict):
+        http_options = HTTPOptions.parse_obj(http_options)
+    if http_options is None:
+        http_options = HTTPOptions(
+            host=http_host, port=http_port, middlewares=http_middlewares)
+
     controller = ServeController.options(
         name=controller_name,
         lifetime="detached" if detached else None,
         max_restarts=-1,
         max_task_retries=-1,
+        # Pin Serve controller on the head node.
+        resources={
+            get_current_node_resource_key(): 0.01
+        },
     ).remote(
         controller_name,
-        http_host,
-        http_port,
-        http_middlewares,
-        detached=detached)
+        http_options,
+        detached=detached,
+    )
 
-    if http_host is not None:
-        futures = []
-        for node_id in ray.state.node_ids():
-            future = block_until_http_ready.options(
-                num_cpus=0, resources={
-                    node_id: 0.01
-                }).remote(
-                    "http://{}:{}/-/routes".format(http_host, http_port),
-                    timeout=HTTP_PROXY_TIMEOUT)
-            futures.append(future)
-        ray.get(futures)
+    proxy_handles = ray.get(controller.get_http_proxies.remote())
+    if len(proxy_handles) > 0:
+        try:
+            ray.get(
+                [handle.ready.remote() for handle in proxy_handles.values()],
+                timeout=HTTP_PROXY_TIMEOUT,
+            )
+        except ray.exceptions.GetTimeoutError:
+            raise TimeoutError(
+                "HTTP proxies not available after {HTTP_PROXY_TIMEOUT}s.")
 
     return Client(controller, controller_name, detached=detached)
 
@@ -464,12 +647,12 @@ def connect() -> Client:
     if not ray.is_initialized():
         ray.init()
 
-    # When running inside of a backend, _INTERNAL_CONTROLLER_NAME is set to
+    # When running inside of a backend, _INTERNAL_REPLICA_CONTEXT is set to
     # ensure that the correct instance is connected to.
-    if _INTERNAL_CONTROLLER_NAME is None:
+    if _INTERNAL_REPLICA_CONTEXT is None:
         controller_name = SERVE_CONTROLLER_NAME
     else:
-        controller_name = _INTERNAL_CONTROLLER_NAME
+        controller_name = _INTERNAL_REPLICA_CONTEXT.controller_name
 
     # Try to get serve controller if it exists
     try:
@@ -483,6 +666,37 @@ def connect() -> Client:
     return Client(controller, controller_name, detached=True)
 
 
+def get_current_backend_tag() -> BackendTag:
+    """When called from within a backend, return its backend tag.
+
+    Raises:
+        RayServeException if not called from within a Ray Serve backend.
+    """
+    if _INTERNAL_REPLICA_CONTEXT is None:
+        raise RayServeException("`serve.get_current_backend_tag()`"
+                                "may only be called from within a"
+                                "Ray Serve backend.")
+    else:
+        return _INTERNAL_REPLICA_CONTEXT.backend_tag
+
+
+def get_current_replica_tag() -> ReplicaTag:
+    """When called from within a backend, return its replica tag.
+
+    A replica tag uniquely identifies a single replica (a process)
+    for a Ray Serve backend.
+
+    Raises:
+        RayServeException if not called from within a Ray Serve backend.
+    """
+    if _INTERNAL_REPLICA_CONTEXT is None:
+        raise RayServeException("`serve.get_current_replica_tag()`"
+                                "may only be called from within a"
+                                "Ray Serve backend.")
+    else:
+        return _INTERNAL_REPLICA_CONTEXT.replica_tag
+
+
 def accept_batch(f: Callable) -> Callable:
     """Annotation to mark that a serving function accepts batches of requests.
 
